@@ -1,3 +1,6 @@
+#define _POSIX_C_SOURCE 200112L
+#define _XOPEN_SOURCE 600
+
 #include <assert.h>
 #include <stdint.h>
 #include <inttypes.h>
@@ -8,6 +11,7 @@
 #include <fcntl.h>
 #include <unistd.h>
 #include <err.h>
+#include <sys/time.h>
 
 #include "utils.h"
 
@@ -36,7 +40,7 @@ static void validate_file(const char *path, int number,
 	uint64_t *ptr_overwritten, uint64_t *ptr_size, int *ptr_read_all,
 	struct timeval *ptr_dt, int progress)
 {
-	char full_fn[PATH_MAX];
+	char *full_fn;
 	const char *filename;
 	uint8_t sector[SECTOR_SIZE], *p, *ptr_end;
 	FILE *f;
@@ -51,10 +55,11 @@ static void validate_file(const char *path, int number,
 
 	*ptr_ok = *ptr_corrupted = *ptr_changed = *ptr_overwritten = 0;
 
-	full_fn_from_number(full_fn, &filename, path, number);
+	full_fn = full_fn_from_number(&filename, path, number);
+	assert(full_fn);
 	printf("Validating file %s ... %s", filename, progress ? BLANK : "");
 	fflush(stdout);
-#ifdef CYGWIN
+#ifdef __CYGWIN__
 	/* We don't need write access, but some kernels require that
 	 * the file descriptor passed to fdatasync(2) to be writable.
 	 */
@@ -96,7 +101,7 @@ static void validate_file(const char *path, int number,
 		for (; error_count <= TOLERANCE && p < ptr_end;
 			p += sizeof(rn)) {
 			rn = random_number(rn);
-			if (rn != *((typeof(rn) *) p))
+			if (rn != *((__typeof__(rn) *) p))
 				error_count++;
 		}
 
@@ -132,7 +137,6 @@ static void validate_file(const char *path, int number,
 
 	*ptr_read_all = feof(f);
 	*ptr_size = ftell(f);
-	assert(*ptr_size >= 0);
 
 	PRINT_STATUS(progress ? CLEAR : "");
 	if (!*ptr_read_all) {
@@ -143,6 +147,7 @@ static void validate_file(const char *path, int number,
 	printf("\n");
 
 	fclose(f);
+	free(full_fn);
 }
 
 static void report(const char *prefix, uint64_t i)
@@ -159,8 +164,8 @@ static inline double dt_to_s(struct timeval *dt)
 	return ret > 0 ? ret : 1;
 }
 
-static void iterate_files(const char *path, const int *files, int start_at,
-	int progress)
+static void iterate_files(const char *path, const long *files,
+	long start_at, long end_at, int progress)
 {
 	uint64_t tot_ok, tot_corrupted, tot_changed, tot_overwritten, tot_size;
 	struct timeval tot_dt = { .tv_sec = 0, .tv_usec = 0 };
@@ -181,10 +186,12 @@ static void iterate_files(const char *path, const int *files, int start_at,
 
 		or_missing_file = or_missing_file || (*files != number);
 		for (; number < *files; number++) {
-			char full_fn[PATH_MAX];
+			char *full_fn;
 			const char *filename;
-			full_fn_from_number(full_fn, &filename, "", number);
+			full_fn = full_fn_from_number(&filename, "", number);
+			assert(full_fn);
 			printf("Missing file %s\n", filename);
+			free(full_fn);
 		}
 		number++;
 
@@ -202,16 +209,18 @@ static void iterate_files(const char *path, const int *files, int start_at,
 	assert(tot_size / SECTOR_SIZE ==
 		(tot_ok + tot_corrupted + tot_changed + tot_overwritten));
 
+	/* Notice that not reporting `missing' files after the last file
+	 * in @files is important since @end_at could be very large.
+	 */
+
 	report("\n  Data OK:", tot_ok);
 	report("Data LOST:", tot_corrupted + tot_changed + tot_overwritten);
 	report("\t       Corrupted:", tot_corrupted);
 	report("\tSlightly changed:", tot_changed);
 	report("\t     Overwritten:", tot_overwritten);
 	if (or_missing_file)
-		printf("WARNING: Not all F3 files are available\n");
-	if (start_at > 0)
-		printf("WARNING: Not all F3 files were tested due to parameter %s%i\n",
-		START_AT_TEXT, start_at + 1);
+		printf("WARNING: Not all F3 files in the range %li to %li are available\n",
+			start_at + 1, end_at + 1);
 	if (!and_read_all)
 		printf("WARNING: Not all data was read due to I/O error(s)\n");
 
@@ -221,39 +230,22 @@ static void iterate_files(const char *path, const int *files, int start_at,
 	printf("Average reading speed: %.2f %s/s\n", read_speed, unit);
 }
 
-int main(int argc, char *argv[])
+int main(int argc, char **argv)
 {
-	int start_at;
+	long start_at, end_at;
 	const char *path;
-	const int *files;
+	const long *files;
 	int progress;
+	int rc;
 
-	switch (argc) {
-	case 2:
-		start_at = 0;
-		path = argv[1];
-		break;
+	rc = parse_args("read", argc, argv, &start_at, &end_at, &path);
+	if (rc)
+		return rc;
 
-	case 3:
-		start_at = parse_start_at_param(argv[1]);
-		if (start_at < 0)
-			goto error;
-		path = argv[2];
-		break;
-
-	default:
-		goto error;
-	}
-
-	files = ls_my_files(path, start_at);
+	files = ls_my_files(path, start_at, end_at);
 	/* If stdout isn't a terminal, supress progress. */
 	progress = isatty(STDOUT_FILENO);
-	iterate_files(path, files, start_at, progress);
+	iterate_files(path, files, start_at, end_at, progress);
 	free((void *)files);
 	return 0;
-
-error:
-	print_header(stderr, "read");
-	fprintf(stderr, "Usage: f3read [%sNUM] <PATH>\n", START_AT_TEXT);
-	return 1;
 }
