@@ -1,3 +1,6 @@
+#define _POSIX_C_SOURCE 200112L
+#define _XOPEN_SOURCE 600
+
 #include <assert.h>
 #include <stdint.h>
 #include <string.h>
@@ -7,10 +10,10 @@
 #include <sys/stat.h>
 #include <fcntl.h>
 #include <sys/statvfs.h>
+#include <sys/time.h>
 #include <errno.h>
 #include <unistd.h>
 #include <err.h>
-#include <alloca.h>
 #include <math.h>
 
 #include "utils.h"
@@ -333,13 +336,13 @@ static inline void end_measurement(int fd, struct flow *fw)
 	fflush(stdout);
 }
 
-static int create_and_fill_file(const char *path, int number, size_t size,
+static int create_and_fill_file(const char *path, long number, size_t size,
 	struct flow *fw)
 {
-	char full_fn[PATH_MAX];
+	char *full_fn;
 	const char *filename;
 	int fd, fine;
-	void *buf;
+	char buf[fw->block_size];
 	size_t remaining;
 	uint64_t offset;
 	ssize_t written;
@@ -348,24 +351,22 @@ static int create_and_fill_file(const char *path, int number, size_t size,
 	assert(size % fw->block_size == 0);
 
 	/* Create the file. */
-	
-	full_fn_from_number(full_fn, &filename, path, number);
+
+	fine = 0;
+	full_fn = full_fn_from_number(&filename, path, number);
+	assert(full_fn);
 	printf("Creating file %s ... ", filename);
 	fflush(stdout);
 	fd = open(full_fn, O_CREAT | O_WRONLY | O_TRUNC, S_IRUSR | S_IWUSR);
 	if (fd < 0) {
 		if (errno == ENOSPC) {
 			printf("No space left.\n");
-			return 0;
+			goto out;
 		}
 		err(errno, "Can't create file %s", full_fn);
 	}
 	assert(fd >= 0);
 
-	/* Obtain the buffer. */
-	buf = alloca(fw->block_size);
-	assert(buf);
-
 	/* Write content. */
 	fine = 1;
 	offset = (uint64_t)number * GIGABYTES;
@@ -388,8 +389,11 @@ static int create_and_fill_file(const char *path, int number, size_t size,
 	assert(!fine || remaining == 0);
 	end_measurement(fd, fw);
 	close(fd);
-	
+
 	printf("OK!\n");
+
+out:
+	free(full_fn);
 	return fine;
 }
 
@@ -407,11 +411,11 @@ static inline void pr_freespace(uint64_t fs)
 	printf("Free space: %.2f %s\n", f, unit);
 }
 
-static int fill_fs(const char *path, int start_at, int progress)
+static int fill_fs(const char *path, long start_at, long end_at, int progress)
 {
 	uint64_t free_space;
 	struct flow fw;
-	int i, fine;
+	long i;
 
 	free_space = get_freespace(path);
 	pr_freespace(free_space);
@@ -420,13 +424,17 @@ static int fill_fs(const char *path, int start_at, int progress)
 		return 1;
 	}
 
+	/* If the amount of data to write is less than the space available,
+	 * update @free_space to improve estimate of time to finish.
+	 */
+	i = end_at - start_at + 1;
+	if (i > 0 && (uint64_t)i <= (free_space >> 30))
+		free_space = (uint64_t)i << 30;
+
 	init_flow(&fw, free_space, progress);
-	i = start_at;
-	fine = 1;
-	do {
-		fine = create_and_fill_file(path, i, GIGABYTES, &fw);
-		i++;
-	} while (fine);
+	for (i = start_at; i <= end_at; i++)
+		if (!create_and_fill_file(path, i, GIGABYTES, &fw))
+			break;
 
 	/* Final report. */
 	pr_freespace(get_freespace(path));
@@ -441,52 +449,37 @@ static int fill_fs(const char *path, int start_at, int progress)
 	return 0;
 }
 
-static void unlink_old_files(const char *path, int start_at)
+static void unlink_old_files(const char *path, long start_at, long end_at)
 {
-	const int *files = ls_my_files(path, start_at);
-	const int *number = files;
+	const long *files = ls_my_files(path, start_at, end_at);
+	const long *number = files;
 	while (*number >= 0) {
-		char full_fn[PATH_MAX];
+		char *full_fn;
 		const char *filename;
-		full_fn_from_number(full_fn, &filename, path, *number);
+		full_fn = full_fn_from_number(&filename, path, *number);
+		assert(full_fn);
 		printf("Removing old file %s ...\n", filename);
 		if (unlink(full_fn))
 			err(errno, "Can't remove file %s", full_fn);
 		number++;
+		free(full_fn);
 	}
 	free((void *)files);
 }
 
-int main(int argc, char *argv[])
+int main(int argc, char **argv)
 {
-	int start_at;
+	long start_at, end_at;
 	const char *path;
 	int progress;
+	int rc;
 
-	switch (argc) {
-	case 2:
-		start_at = 0;
-		path = argv[1];
-		break;
+	rc = parse_args("write", argc, argv, &start_at, &end_at, &path);
+	if (rc)
+		return rc;
 
-	case 3:
-		start_at = parse_start_at_param(argv[1]);
-		if (start_at < 0)
-			goto error;
-		path = argv[2];
-		break;
-
-	default:
-		goto error;
-	}
-
-	unlink_old_files(path, start_at);
+	unlink_old_files(path, start_at, end_at);
 	/* If stdout isn't a terminal, supress progress. */
 	progress = isatty(STDOUT_FILENO);
-	return fill_fs(path, start_at, progress);
-
-error:
-	print_header(stderr, "write");
-	fprintf(stderr, "Usage: f3write [%sNUM] <PATH>\n", START_AT_TEXT);
-	return 1;
+	return fill_fs(path, start_at, end_at, progress);
 }
